package kas

import (
	"strconv"
	"time"

	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/module/modshared"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/tool/grpctool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/tool/logz"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/tool/retry"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/tunnel/tunserver"
	"go.uber.org/zap"
	"google.golang.org/grpc"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/metadata"
	"google.golang.org/grpc/status"
)

var (
	_ tunserver.GatewayFinderFactory = (*GatewayFinderFactory)(nil)
)

type GatewayFinderFactory struct {
	KASPool               grpctool.PoolInterface
	GatewayQuerier        tunserver.PollingGatewayURLQuerier
	OwnPrivateAPIURL      string
	PollConfig            retry.PollConfigFactory
	TryNewGatewayInterval time.Duration
}

func (f *GatewayFinderFactory) GatewayFinderForStream(stream grpc.ServerStream, rpcAPI modshared.RPCAPI) (tunserver.GatewayFinder, *zap.Logger, int64, error) {
	ctx := stream.Context()
	md, _ := metadata.FromIncomingContext(ctx)
	agentID, err := agentIDFromMeta(md)
	if err != nil {
		return nil, nil, 0, err // returns gRPC status error
	}

	log := rpcAPI.Log().With(logz.AgentID(agentID))
	gf := tunserver.NewGatewayFinder(
		log,
		f.KASPool,
		f.GatewayQuerier,
		rpcAPI,
		grpc.ServerTransportStreamFromContext(ctx).Method(),
		f.OwnPrivateAPIURL,
		agentID,
		metadata.NewOutgoingContext(ctx, md),
		f.PollConfig,
		f.TryNewGatewayInterval,
	)
	return gf, log, agentID, nil
}

// agentIDFromMeta returns agent ID or an gRPC status error.
func agentIDFromMeta(md metadata.MD) (int64, error) {
	val := md.Get(tunserver.RoutingAgentIDMetadataKey)
	if len(val) != 1 {
		return 0, status.Errorf(codes.InvalidArgument, "Expecting a single %s, got %d", tunserver.RoutingAgentIDMetadataKey, len(val))
	}
	agentID, err := strconv.ParseInt(val[0], 10, 64)
	if err != nil {
		return 0, status.Errorf(codes.InvalidArgument, "Invalid %s", tunserver.RoutingAgentIDMetadataKey)
	}
	return agentID, nil
}
